/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// quakedef.h -- primary header for server

#define	QUAKE_GAME			// as opposed to utilities

//define	PARANOID			// speed sapping error checking

#ifdef _WIN32
#pragma warning( disable : 4244 4127 4201 4214 4514 4305 4115 4018)
#endif

#include <math.h>
#include <string.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <setjmp.h>
#include <ctype.h>
//#include <ust/marker.h>

#include "bothdefs.h"

#include "common.h"
#include "bspfile.h"
#include "sys.h"
#include "zone.h"
#include "mathlib.h"

#include "cvar.h"
#include "net.h"
#include "protocol.h"
#include "cmd.h"
#include "model.h"
#include "crc.h"
#include "progs.h"

#include "server.h"
#include "world.h"
#include "pmove.h"
#include "qsort.h"

//=============================================================================

// the host system specifies the base of the directory tree, the
// command line parms passed to the program, and the amount of memory
// available for the program to use

typedef struct
{
	char	*basedir;
	char	*cachedir;		// for development over ISDN lines
	int		argc;
	char	**argv;
	void	*membase;
	int		memsize;
} quakeparms_t;


//=============================================================================

//
// host
//
extern	quakeparms_t host_parms;

extern	cvar_t		sys_nostdout;
extern	cvar_t		developer;

extern	qboolean	host_initialized;		// true if into command execution
extern	double		host_frametime;
extern	double		realtime;			// not bounded in any way, changed at
										// start of every frame, never reset

void SV_Error (char *error, ...);
void SV_Init (quakeparms_t *parms);

void Con_Printf (char *fmt, ...);
void Con_DPrintf (char *fmt, ...);


//=============================================================================

// Globals and function prototypes related to the parallel server execution.

extern int serverForked;
extern int processorID;
extern int taskSamples[180][MAX_EDICTS];

extern cvar_t par_debugmsgs;        // Whether we should display debug messages.
extern cvar_t par_benchmark;        // Whether we should benchmark server performance.
extern cvar_t par_pmovesize;        // The size of the bounding box used in pmove calculation.
extern cvar_t par_numprocs;         // The number of child processes to use for execution.
extern cvar_t par_edictfix;         // Propagate edict creations into the coordinator thread.
extern cvar_t par_forceretouch;     // Manually force entities into the areanode tree.
extern cvar_t par_sortingalgo;      // Select between different sorting algorithms.
extern cvar_t par_forceaffinity;    // Manually force process affinity or leave to the OS scheduler.
extern cvar_t par_deferredresponse; // Send out responses from coordinator thread instead of children.
extern cvar_t par_loadbalance;      // Selects between different load balancing strategies.
extern cvar_t par_clienttarget;		// The number of clients required before benchmarking begins.
extern cvar_t par_frametarget;      //
extern cvar_t par_groupinfo;		// Print out contents of workgroups each frame.
extern cvar_t par_sleeptime;
extern cvar_t par_imbalance;
extern cvar_t par_huffman_trans;
extern cvar_t par_sendproc;
extern cvar_t par_packetsort;

#define NPROCS 6

void SV_Fork(void);
void SV_Wait(void);
void SV_ConstructWorkGroups(void);
void SV_LoadBalance( void );
int  SV_ValidateWorkGroup(client_t *);

typedef struct {
  double 		start_time;
  unsigned int 	frames;
  unsigned int 	total_groups;
  unsigned int 	tam_grupos;
  unsigned int 	packets[NPROCS];
  unsigned int 	responses[NPROCS];
  unsigned int 	weight[NPROCS];
  unsigned int	edict_propagations[NPROCS];
  double 		total_wait_time;
} statistics_t;

extern statistics_t *stats;

