using System;
using System.Text;
using System.Net.Sockets;
using System.Threading;
using System.Net;
using System.Collections.Generic;

namespace ClientLauncherServer
{
	public class ListenServer
	{
		private TcpListener tcpListener;
		private Thread listenThread;
		private List<TcpClient> clients = new List<TcpClient>();
		private List<TcpClient> reserves = new List<TcpClient>();
		
		public ListenServer ()
		{
		}
		
		public void Start()
		{
			this.tcpListener = new TcpListener(IPAddress.Any, 3000);
			this.listenThread = new Thread(new ThreadStart(ListenForClients));
			this.listenThread.Start();
		}
		
		private void ListenForClients()
		{
			this.tcpListener.Start();
			
			while (true)
			{
				// Blocks until a client has connected to the server
				TcpClient client = this.tcpListener.AcceptTcpClient();
				client.Client.SetSocketOption(SocketOptionLevel.Socket,SocketOptionName.KeepAlive,true);
				client.Client.ReceiveTimeout = 180000;
				

				// Create a thread to handle communication with connected client
				Thread clientThread = new Thread(new ParameterizedThreadStart(HandleClient));
				clientThread.Start(client);
			}
		}
	
		private void HandleClient(object client)
		{
			TcpClient tcpClient = (TcpClient)client;
			NetworkStream clientStream = tcpClient.GetStream();

			byte[] message = new byte[4096];
			int bytesRead;
		
			while (true)
			{
				bytesRead = 0;

				try
				{
					// Blocks until a client sends a message
					bytesRead = clientStream.Read(message, 0, 4096);
				}
				catch
				{
					// A socket error has occured
					break;
				}

				if (bytesRead == 0)
				{ 
					break;
				}

				// Message has successfully been received
				ASCIIEncoding encoder = new ASCIIEncoding();
				//Console.WriteLine(encoder.GetString(message, 0, bytesRead));
				this.HandleClientMessage(tcpClient,encoder.GetString(message, 0, bytesRead));
			}
			
			lock( clients ) {
				ClientDisconnect(tcpClient);
			} 
		}
		
		private void HandleClientMessage(TcpClient client, string message)
		{
			if( message == "PING regular" ) {
				this.SendClientMessage(client,"PONG ok");
				
				lock( clients ) {
					IPEndPoint ip = (IPEndPoint)client.Client.RemoteEndPoint;
					Console.WriteLine("Successfully authenticated client #{0} ({1}).",
					                  clients.Count,IPAddress.Parse(ip.Address.ToString()));
					clients.Add(client);
				}
			} else if( message == "PING reserve" ) {
				this.SendClientMessage(client,"PONG ok");
				
				lock( clients ) {
					IPEndPoint ip = (IPEndPoint)client.Client.RemoteEndPoint;
					Console.WriteLine("Successfully authenticated reserved client #{0} ({1}).",
					                  reserves.Count,IPAddress.Parse(ip.Address.ToString()));
					reserves.Add(client);
				}
			} else if( message == "exiting" ) {
				ClientDisconnect(client);
			} else if( message == "ping" ) {
				SendClientMessage(client,"pong");
				//Console.WriteLine("Ponged...");
			}
			
			/*NetworkStream  clientStream = tcpClient.GetStream();
			ASCIIEncoding encoder = new ASCIIEncoding();
			byte[] buffer = encoder.GetBytes("Hello Client!");

			clientStream.Write(buffer, 0 , buffer.Length);
			clientStream.Flush();*/
		}
		
		private void SendClientMessage(TcpClient client, string message)
		{
			NetworkStream clientStream = client.GetStream();
            ASCIIEncoding encoder = new ASCIIEncoding();
			byte[] buffer = encoder.GetBytes(message);
			clientStream.Write(buffer, 0, buffer.Length);
            clientStream.Flush();
		}
		
		private void SendAllClientMessage(string message)
		{
			lock( clients ) {
				foreach( TcpClient client in clients ) {
					this.SendClientMessage(client,message);
				}
			}
		}
		
		private void PromoteReserve(TcpClient reserve)
		{
			int id = reserves.IndexOf(reserve);
			
			reserves.Remove(reserve);
			clients.Add(reserve);
			Console.WriteLine("Promoted reserved client #{0} to active client (#{1})",
			                  id,clients.IndexOf(reserve));
		}
		
		private void DemoteReserve(TcpClient reserve)
		{
			int id = clients.IndexOf(reserve);
			
			clients.Remove(reserve);
			reserves.Add(reserve);
			Console.WriteLine("Demoted regular client #{0} to reserved client (#{1})",
			                  id,reserves.IndexOf(reserve));
		}
		
		private void ClientDisconnect( TcpClient client )
		{
			client.Close();
			
			if( clients.Contains(client) ) {
				Console.WriteLine("Removing client #{0}.",clients.IndexOf(client));
				clients.Remove(client);
			} else if( reserves.Contains(client) ) {
				Console.WriteLine("Removing reserved client #{0}.",reserves.IndexOf(client));
				reserves.Remove(client);
			} else 
				Console.WriteLine("Removing untracked client.");
		}

		public void ParseInput(string stringInput)
		{
			string[] args = stringInput.Split(' ');
			
			lock( clients ) {
				if( clients.Count == 0 ) {
					Console.WriteLine("No clients to broadcast to!");
					return;
				} else if( stringInput.Contains("launch") ) {
					if( args.Length == 2 ) {
						StringBuilder cmd = new StringBuilder("launch ");
						cmd.Append(args[1]);
						SendAllClientMessage(cmd.ToString());
					} else if( args.Length == 3 ) {
						if( args[1] == "extra" ) {
							int i = 0;
							int client = 0;
							int num = Int32.Parse(args[2]);
							while( i < num ) {
								SendClientMessage(clients[client],"launch 1");
								++client;
								++i;
								
								if( client >= clients.Count )
									client = 0;
							}
						}
					} else Console.WriteLine("Too few arguments!");
				} else if( stringInput == "kill" ) {
					SendAllClientMessage("kill all");
				} else if( stringInput == "list reserves" ) {
					int i = -1;
					foreach( TcpClient reserve in reserves ) {
						IPEndPoint ip = (IPEndPoint)reserve.Client.RemoteEndPoint;
						Console.WriteLine("Reserve #{0} ({1}).",
					                  ++i,IPAddress.Parse(ip.Address.ToString()));
					}
				} else if( stringInput == "list clients" ) {
					int i = -1;
					foreach( TcpClient client in clients ) {
						IPEndPoint ip = (IPEndPoint)client.Client.RemoteEndPoint;
						Console.WriteLine("Client #{0} ({1}).",
					                  ++i,IPAddress.Parse(ip.Address.ToString()));
					}
				} else if( stringInput.Contains("promote") ) {
					if( args.Length == 2 ) {
						if( reserves.Count == 0 ) {
							Console.WriteLine("No reserved clients available!");
							return;
						}
						
						int index = Int32.Parse(args[1]);
						if( index + 1 <= reserves.Count || index < 0 )
							PromoteReserve(reserves[index]);
						else Console.WriteLine("Invalid reserve ID!");
					} else Console.WriteLine("Too few arguments!");
				} else if( stringInput.Contains("demote") ) {
					if( args.Length == 2 ) {
						int index = Int32.Parse(args[1]);
						if( index + 1 <= clients.Count || index < 0 )
							DemoteReserve(clients[index]);
						else Console.WriteLine("Invalid reserve ID!");
					} else Console.WriteLine("Too few arguments!");
				} else if( stringInput == "remove all" ) {
					foreach( TcpClient client in clients ) {
						Console.WriteLine("Removing client #{0}.",clients.IndexOf(client));
						this.SendClientMessage(client,"quit");
						client.Close();
					}
					
					clients.Clear();
				} else Console.WriteLine("Unknown command!");
			}
		}
	}
}
