
using System;
using ZedGraph;
using System.Drawing;
using System.Drawing.Imaging;
using System.Threading;

namespace DistribAnalyser
{
	public class GraphTask
	{
		public GraphTask (int[,] workGroups, int start, int end, int numThreads, string prefix)
		{
			_workGroups = workGroups;
			_start = start;
			_end = end;
			_numThreads = numThreads;
			_prefix = prefix;
			_numGraphs = start;
		}
		
		public void Start()
		{
			_thread = new Thread(new ThreadStart(LoadBalance));
			_thread.Start();
		}
		
		public void Join()
		{
			_thread.Join();
		}
		
		public void LoadBalance()
        {
            int j = -1;
            int[] threadWeight = new int[_numThreads];
            int lessWeight = 0;

            for (int i = _start; i < _end; ++i)
            {
                j = -1;
                threadWeight[0] = 1;

                while (_workGroups[i, ++j] != 0) {}
				for( int l = j; l >= 0; --l )
                {
                    lessWeight = 0;

                    for (int k = 0; k < _numThreads; ++k)
                    {
                        if (threadWeight[k] < threadWeight[lessWeight])
                        {
                            lessWeight = k;
                        }
                    }

                    threadWeight[lessWeight] += _workGroups[i,l];
                }

                this.OutputChart(threadWeight);

                for (int k = 0; k < _numThreads; ++k)
                {
                    threadWeight[k] = 0;
                }
            }
        }
		
		private void OutputChart(int[] threads)
        {
			GraphPane graphPane = new GraphPane(new RectangleF(0, 0, 640, 720),
                string.Format("Workload Distribution for {0} Threads (Sample {1} of {2})", _numThreads, _numGraphs + 1, 180),
                "", "Thread Weight (No. Entities)");
            PointPairList[] ppList = new PointPairList[_numThreads];

            for (int i = 0; i < _numThreads; ++i)
            {
                Color colour = Color.Empty;
                switch (i)
                {
                    case 0: colour = Color.Red; break;
                    case 1: colour = Color.Yellow; break;
                    case 2: colour = Color.Pink; break;
                    case 3: colour = Color.Green; break;
                    case 4: colour = Color.Orange; break;
                    case 5: colour = Color.Purple; break;
                    case 6: colour = Color.Blue; break;
                    case 7: colour = Color.Orange; break;
                }

                ppList[i] = new PointPairList();
                ppList[i].Add(0, (double)threads[i]);
                graphPane.AddBar(string.Format("Thread {0}",i + 1), ppList[i], colour);
            }

            graphPane.Chart.Fill = new Fill(Color.White,Color.DarkGray, 45.0F);

            Bitmap bm = new Bitmap(1, 1);
            using (Graphics g = Graphics.FromImage(bm))
                graphPane.AxisChange(g);

            graphPane.YAxis.Scale.Min = 0;
            graphPane.YAxis.Scale.Max = 75;
            graphPane.YAxis.Scale.MinorStep = 5.0;
            graphPane.YAxis.Scale.MajorStep = 10.0;
            graphPane.XAxis.MajorTic.Size = 0.0f;
            graphPane.XAxis.MinorTic.Size = 0.0f;
            graphPane.BarSettings.ClusterScaleWidth = 1;
            graphPane.BarSettings.MinBarGap = 0.2f;
            BarItem.CreateBarLabels(graphPane, false, "f0");
            graphPane.GetImage().Save(string.Format("{0}{1}.png", _prefix, _numGraphs), ImageFormat.Png);

            Console.WriteLine("Outputing chart {0} of {1}...", _numGraphs + 1, 180);
            ++_numGraphs;
        }
		
		Thread _thread;
		int[,] _workGroups;
		int _start;
		int _end;
		int _numThreads;
		int _numGraphs;
		string _prefix;
	}
}
